@@ -3,13 +3,10 @@
 use std::collections::HashMap;
 
 use proc_macro::TokenStream;
-use proc_macro2::Group;
 use quote::format_ident;
 use quote::quote;
 use quote::quote_spanned;
 use quote::ToTokens;
-use syn::parenthesized;
-use syn::parse::Parse;
 use syn::parse_macro_input;
 use syn::spanned::Spanned;
 use syn::token::Brace;
@@ -18,20 +15,16 @@ use syn::token::Lt;
 use syn::token::Trait;
 use syn::AngleBracketedGenericArguments;
 use syn::Block;
-use syn::Error;
 use syn::Expr;
 use syn::GenericArgument;
 use syn::GenericParam;
-use syn::Ident;
 use syn::ImplItemMethod;
 use syn::ItemTrait;
-use syn::LitInt;
 use syn::Path;
 use syn::PathArguments;
 use syn::PathSegment;
 use syn::Signature;
 use syn::Stmt;
-use syn::Token;
 use syn::TraitBound;
 use syn::TraitItem;
 use syn::TraitItemMethod;
@@ -40,22 +33,23 @@ use syn::TypeParam;
 use syn::TypeParamBound;
 use syn::TypePath;
 use syn::Visibility;
-use syn_utils::TypeOrPath;
 
 use crate::parse_assoc_type::parse_assoc_type;
 use crate::parse_assoc_type::AssocTypeError;
-use crate::parse_trait_sig::parse_trait_signature;
-use crate::parse_trait_sig::MethodError;
-use crate::parse_trait_sig::SignatureChanges;
-use crate::parse_trait_sig::TypeTransform;
+use crate::parse_attrs::TraitAttrs;
 use crate::syn_utils::iter_path;
 use crate::syn_utils::trait_bounds;
+use crate::trait_sig::convert_trait_signature;
+use crate::trait_sig::MethodError;
+use crate::trait_sig::SignatureChanges;
+use crate::trait_sig::TypeTransform;
 use crate::transform::TransformError;
 use crate::transform::TypeConverter;
 
 mod parse_assoc_type;
-mod parse_trait_sig;
+mod parse_attrs;
 mod syn_utils;
+mod trait_sig;
 mod transform;
 
 macro_rules! abort {
@@ -66,91 +60,48 @@ macro_rules! abort {
     }};
 }
 
-struct Collection {
-    id: Ident,
-    count: usize,
-}
-
-impl Parse for Collection {
-    fn parse(input: syn::parse::ParseStream) -> syn::Result<Self> {
-        let inner;
-        parenthesized!(inner in input);
-
-        let id: Ident = inner.parse()?;
-        let _: Token![,] = inner.parse()?;
-        let count_lit: LitInt = inner.parse()?;
-        let count: usize = count_lit.base10_parse()?;
-        if count < 1 {
-            return Err(Error::new(
-                count_lit.span(),
-                "number of type parameters must be >= 1",
-            ));
-        }
-        Ok(Self { id, count })
-    }
-}
-
 #[proc_macro_attribute]
 pub fn dynamize(_attr: TokenStream, input: TokenStream) -> TokenStream {
     let mut original_trait = parse_macro_input!(input as ItemTrait);
     assert!(original_trait.auto_token.is_none());
 
-    let original_trait_name = original_trait.ident.clone();
-
-    let mut objectifiable_methods: Vec<(Signature, SignatureChanges)> = Vec::new();
-
-    let mut type_converter = TypeConverter::default();
-
-    let mut blanket_impl_attrs = Vec::new();
-    let mut dyn_trait_attrs = Vec::new();
-
-    // FUTURE: use Vec::drain_filter once it's stable
-    let mut i = 0;
-    while i < original_trait.attrs.len() {
-        if original_trait.attrs[i].path.is_ident("blanket_impl_attr") {
-            let attr = original_trait.attrs.remove(i);
-            let group: Group = match syn::parse2(attr.tokens) {
-                Ok(g) => g,
-                Err(err) => {
-                    return abort!(
-                        err.span(),
-                        "expected parenthesis: #[blanket_impl_attr(...)]"
-                    )
-                }
-            };
-            let tokens = group.stream();
-            blanket_impl_attrs.push(quote! {#[#tokens]});
-        } else if original_trait.attrs[i].path.is_ident("dyn_trait_attr") {
-            let attr = original_trait.attrs.remove(i);
-            let group: Group = match syn::parse2(attr.tokens) {
-                Ok(g) => g,
-                Err(err) => {
-                    return abort!(err.span(), "expected parenthesis: #[dyn_trait_attr(...)]")
-                }
-            };
-            let tokens = group.stream();
-            dyn_trait_attrs.push(quote! {#[#tokens]});
-        } else if original_trait.attrs[i].path.is_ident("collection") {
-            let attr = original_trait.attrs.remove(i);
-            let tokens = attr.tokens.into();
-            let coll = parse_macro_input!(tokens as Collection);
-
-            if type_converter
-                .collections
-                .insert(coll.id.clone(), coll.count)
-                .is_some()
-            {
-                return abort!(
-                    coll.id.span(),
-                    "collection `{}` is defined multiple times for this trait",
-                    coll.id
-                );
-            }
-        } else {
-            i += 1;
+    // just to provide a better error message
+    for bound in trait_bounds(&original_trait.supertraits) {
+        if let Some(assoc_type) = iter_path(&bound.path)
+            .filter_map(filter_map_assoc_paths)
+            .next()
+        {
+            return abort!(
+                assoc_type.span(),
+                "dynamize does not support associated types in supertraits"
+            );
         }
     }
 
+    let method_attrs = match TraitAttrs::parse(&mut original_trait.attrs) {
+        Ok(attrs) => attrs,
+        Err(err) => return err.to_compile_error().into(),
+    };
+
+    for dyn_supertrait in &method_attrs.dynamized_supertraits {
+        if !trait_bounds(&original_trait.supertraits)
+            .any(|t| t.path.segments[0].ident == *dyn_supertrait)
+        {
+            return abort!(
+                dyn_supertrait.span(),
+                "this trait definition has no such supertrait"
+            );
+        }
+    }
+
+    let mut type_converter = TypeConverter {
+        trait_ident: original_trait.ident.clone(),
+        assoc_type_conversions: HashMap::new(),
+        collections: method_attrs.collections,
+        type_conversions: method_attrs.type_conversions,
+        used_conversions: Default::default(),
+    };
+
     for item in &original_trait.items {
         if let TraitItem::Type(assoc_type) = item {
             match parse_assoc_type(assoc_type) {
@@ -173,10 +124,13 @@ pub fn dynamize(_attr: TokenStream, input: TokenStream) -> TokenStream {
         }
     }
 
-    for item in &original_trait.items {
+    let mut objectifiable_methods: Vec<(Signature, SignatureChanges)> = Vec::new();
+
+    for item in &mut original_trait.items {
         if let TraitItem::Method(method) = item {
             let mut signature = method.sig.clone();
-            match parse_trait_signature(&mut signature, &type_converter) {
+
+            match convert_trait_signature(&mut signature, &mut type_converter) {
                 Ok(parsed_method) => objectifiable_methods.push((signature, parsed_method)),
                 Err((span, err)) => match err {
                     MethodError::NonDispatchableMethod => continue,
@@ -216,6 +170,12 @@ pub fn dynamize(_attr: TokenStream, input: TokenStream) -> TokenStream {
                             n
                         )
                     }
+                    MethodError::Transform(TransformError::QualifiedAssociatedType) => {
+                        return abort!(span, "dynamize does not support qualified associated types")
+                    }
+                    MethodError::Transform(TransformError::SelfQualifiedAsOtherTrait) => {
+                        return abort!(span, "dynamize does not know how to convert this type (you can tell it with #[convert])")
+                    }
                     MethodError::UnconvertedAssocType => {
                         return abort!(span, "dynamize does not support associated types here")
                     }
@@ -224,6 +184,13 @@ pub fn dynamize(_attr: TokenStream, input: TokenStream) -> TokenStream {
         }
     }
 
+    for ty in type_converter.type_conversions.keys() {
+        if !type_converter.used_conversions.contains(ty) {
+            // FUTURE: relax to warning once proc_macro::Diagnostic is stable
+            return abort!(ty.span(), "unused conversion");
+        }
+    }
+
     let mut method_impls: Vec<ImplItemMethod> = Vec::new();
 
     let mut dyn_trait = ItemTrait {
@@ -239,11 +206,20 @@ pub fn dynamize(_attr: TokenStream, input: TokenStream) -> TokenStream {
         supertraits: original_trait
             .supertraits
             .iter()
-            .filter(|t| match t {
-                TypeParamBound::Trait(t) => !t.path.is_ident("Sized"),
-                TypeParamBound::Lifetime(_) => true,
+            .filter_map(|t| {
+                if let TypeParamBound::Trait(trait_bound) = t {
+                    if trait_bound.path.is_ident("Sized") {
+                        return None;
+                    }
+                    let ident = &trait_bound.path.segments[0].ident;
+                    if method_attrs.dynamized_supertraits.contains(ident) {
+                        let mut bound = trait_bound.clone();
+                        bound.path.segments[0].ident = format_ident!("Dyn{}", ident);
+                        return Some(TypeParamBound::Trait(bound));
+                    }
+                }
+                Some(t.clone())
             })
-            .cloned()
             .collect(),
         brace_token: Brace::default(),
         items: Vec::new(),
@@ -264,6 +240,8 @@ pub fn dynamize(_attr: TokenStream, input: TokenStream) -> TokenStream {
         }
     }
 
+    let original_trait_name = original_trait.ident.clone();
+
     for (signature, parsed_method) in objectifiable_methods {
         let mut new_method = TraitItemMethod {
             attrs: Vec::new(),
@@ -377,6 +355,9 @@ pub fn dynamize(_attr: TokenStream, input: TokenStream) -> TokenStream {
     let dyn_trait_name = &dyn_trait.ident;
     let (impl_generics, ty_generics, where_clause) = dyn_trait.generics.split_for_impl();
 
+    let dyn_trait_attrs = method_attrs.dyn_trait_attrs;
+    let blanket_impl_attrs = method_attrs.blanket_impl_attrs;
+
     let expanded = quote! {
         #original_trait
 
@@ -425,20 +406,30 @@ fn generate_blanket_impl(
     }
 }
 
-fn path_is_assoc_type(path: &Path) -> bool {
-    path.segments[0].ident == "Self"
+fn path_is_assoc_type(path: &TypePath) -> bool {
+    if path.path.segments[0].ident == "Self" {
+        return true;
+    }
+    if let Some(qself) = &path.qself {
+        if let Type::Path(path) = qself.ty.as_ref() {
+            if path.path.segments[0].ident == "Self" {
+                return true;
+            }
+        }
+    }
+    false
 }
 
-fn match_assoc_type(item: TypeOrPath) -> bool {
-    if let TypeOrPath::Path(path) = item {
+fn match_assoc_type(item: &Type) -> bool {
+    if let Type::Path(path) = item {
         return path_is_assoc_type(path);
     }
     false
 }
 
-fn filter_map_assoc_paths(item: TypeOrPath) -> Option<&Path> {
+fn filter_map_assoc_paths(item: &Type) -> Option<&TypePath> {
     match item {
-        TypeOrPath::Path(p) if path_is_assoc_type(p) => Some(p),
+        Type::Path(p) if path_is_assoc_type(p) => Some(p),
         _other => None,
     }
 }
@@ -492,6 +483,11 @@ impl TypeTransform {
                 quote! {#arg.into_iter().map(|(#(#idents),*)| (#(#transforms),*)).collect()}
             }
             TypeTransform::NoOp => arg,
+            TypeTransform::Verbatim(convert) => {
+                let ident = &convert.ident;
+                let block = &convert.block;
+                quote! { {let #ident = #arg; #block }}
+            }
         }
     }
 }
