@@ -1,4 +1,7 @@
-use std::collections::HashMap;
+use std::{
+    collections::{HashMap, HashSet},
+    rc::Rc,
+};
 
 use proc_macro2::Span;
 use quote::quote;
@@ -10,14 +13,17 @@ use syn::{
 use crate::{
     filter_map_assoc_paths, match_assoc_type,
     parse_assoc_type::{BoxType, DestType},
-    parse_trait_sig::{MethodError, TypeTransform},
+    parse_attrs::Convert,
     syn_utils::{iter_path, iter_type, type_arguments_mut},
+    trait_sig::{MethodError, TypeTransform},
 };
 
-#[derive(Default)]
-pub struct TypeConverter<'a> {
-    pub assoc_type_conversions: HashMap<Ident, DestType<'a>>,
+pub struct TypeConverter {
+    pub assoc_type_conversions: HashMap<Ident, DestType>,
     pub collections: HashMap<Ident, usize>,
+    pub trait_ident: Ident,
+    pub type_conversions: HashMap<Type, Rc<Convert>>,
+    pub used_conversions: HashSet<Type>,
 }
 
 #[derive(Debug)]
@@ -26,9 +32,11 @@ pub enum TransformError {
     UnsupportedType,
     ExpectedAtLeastNTypes(usize),
     AssocTypeAfterFirstNTypes(usize, Ident),
+    QualifiedAssociatedType,
+    SelfQualifiedAsOtherTrait,
 }
 
-impl TypeConverter<'_> {
+impl TypeConverter {
     /// A return type of Some(1) means that the type implements
     /// IntoIterator<Item=T1> and FromIterator<T1>
     /// with T1 being its first generic type parameter.
@@ -56,7 +64,15 @@ impl TypeConverter<'_> {
         None
     }
 
-    pub fn convert_type(&self, type_: &mut Type) -> Result<TypeTransform, (Span, TransformError)> {
+    pub fn convert_type(
+        &mut self,
+        type_: &mut Type,
+    ) -> Result<TypeTransform, (Span, TransformError)> {
+        if let Some(conv) = self.type_conversions.get(type_) {
+            self.used_conversions.insert(conv.original_type.clone());
+            *type_ = conv.dest_type.clone();
+            return Ok(TypeTransform::Verbatim(conv.clone()));
+        }
         if !iter_type(type_).any(match_assoc_type) {
             return Ok(TypeTransform::NoOp);
         }
@@ -112,7 +128,35 @@ impl TypeConverter<'_> {
             }
         }
 
-        if let Type::Path(TypePath { path, qself: None }) = type_ {
+        if let Type::Path(TypePath {
+            path,
+            qself: Some(qself),
+        }) = type_
+        {
+            if let Type::Path(self_path) = qself.ty.as_ref() {
+                if self_path.path.segments[0].ident == "Self" {
+                    if !self_path.path.is_ident("Self") {
+                        return Err((self_path.span(), TransformError::QualifiedAssociatedType));
+                    }
+
+                    if qself.position == 1
+                        && path.segments.len() == 2
+                        && path.segments[0].arguments.is_empty()
+                        && path.segments[0].ident == self.trait_ident
+                    {
+                        let ident = &path.segments[1].ident;
+                        let dest_type =
+                            self.assoc_type_conversions.get(ident).ok_or_else(|| {
+                                (ident.span(), TransformError::AssocTypeWithoutDestType)
+                            })?;
+                        *type_ = dest_type.get_dest();
+                        return Ok(dest_type.type_transformation());
+                    }
+
+                    return Err((type_.span(), TransformError::SelfQualifiedAsOtherTrait));
+                }
+            }
+        } else if let Type::Path(TypePath { path, qself: None }) = type_ {
             if path.segments[0].ident == "Self" {
                 if path.segments.len() == 2 {
                     let ident = &path.segments.last().unwrap().ident;
@@ -138,10 +182,10 @@ impl TypeConverter<'_> {
                                     TransformError::ExpectedAtLeastNTypes(type_count),
                                 ));
                             }
-                            for i in type_count..args.len() {
-                                if iter_type(args[i]).any(match_assoc_type) {
+                            for ty in args.iter().skip(type_count) {
+                                if iter_type(ty).any(match_assoc_type) {
                                     return Err((
-                                        args[i].span(),
+                                        ty.span(),
                                         TransformError::AssocTypeAfterFirstNTypes(
                                             type_count,
                                             last_seg.ident.clone(),
@@ -187,7 +231,7 @@ impl TypeConverter<'_> {
 
 pub fn dynamize_function_bounds(
     generics: &mut Generics,
-    type_converter: &TypeConverter,
+    type_converter: &mut TypeConverter,
 ) -> Result<HashMap<Ident, Vec<TypeTransform>>, (Span, MethodError)> {
     let mut type_param_transforms = HashMap::new();
 
@@ -249,7 +293,7 @@ pub fn dynamize_function_bounds(
 
 fn dynamize_trait_bound(
     bound: &mut TraitBound,
-    type_converter: &TypeConverter,
+    type_converter: &mut TypeConverter,
     type_ident: &Ident,
     type_param_transforms: &mut HashMap<Ident, Vec<TypeTransform>>,
 ) -> Result<(), (Span, MethodError)> {
