@@ -35,6 +35,8 @@ fn it_works() {
         #[allow(clippy::type_complexity)]
         fn test10(&self) -> (Self::A, (Self::A, Self::B), Self::B);
 
+        fn test11(&self) -> <Self as MyTrait>::A;
+
         // fn test9(&self) -> &dyn Iterator<Item = Self::A>;
 
         fn mut1(&mut self) -> Self::A;
@@ -280,3 +282,41 @@ trait CustomCollection {
 
     fn test(&self) -> MyCollection<Self::A, Self::A, Self::A>;
 }
+
+#[dynamize::dynamize]
+trait SomeTrait {
+    type Y: Into<String>;
+
+    fn foo(&self) -> Self::Y;
+}
+
+#[dynamize::dynamize]
+#[dynamized(SomeTrait)]
+trait OneTrait: SomeTrait {
+    type X: Into<String>;
+
+    fn test(&self) -> Self::X;
+}
+
+#[dynamize::dynamize]
+#[dynamized(Foo)]
+trait OtherTrait<X>: Foo<X> {}
+
+struct SuperError;
+trait Cache {}
+#[dynamize::dynamize]
+trait Client {
+    type Error: std::error::Error;
+
+    fn get(&self, url: String) -> Result<Vec<u8>, Self::Error>;
+}
+#[dynamize::dynamize]
+#[dynamized(Client)]
+#[convert = |x: <Self as Client>::Error| -> Box<dyn std::error::Error + '_> {Box::new(x) as _}]
+trait ClientWithCache: Client {
+    fn get_with_cache<C: Cache>(
+        &self,
+        url: String,
+        cache: C,
+    ) -> Result<Vec<u8>, <Self as Client>::Error>;
+}
